

//_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/
//_/_/
//_/_/  This file is part of CMSDK version 2.0.1.180405 (05 April 2018)
//_/_/  
//_/_/  developed by
//_/_/  
//_/_/  Communicative Machines Ltd. 
//_/_/    http://cmlabs.com/
//_/_/  Copyright (c) 2018
//_/_/  
//_/_/  in collaboration with 
//_/_/  
//_/_/  Icelandic Institute for Intelligent Machines ses
//_/_/    Menntavegur 1
//_/_/    101 Reykjavik
//_/_/    ICELAND
//_/_/    http://www.iiim.is/
//_/_/
//_/_/  and is released under the following
//_/_/  
//_/_/  EXTENDED BSD License
//_/_/
//_/_/  Redistribution and use in source and binary forms, with or without
//_/_/  modification, is permitted provided that the following conditions 
//_/_/  are met:
//_/_/
//_/_/  - Redistributions of source code must retain the above copyright 
//_/_/    and collaboration notice, this list of conditions and the following 
//_/_/    disclaimer.
//_/_/
//_/_/  - Redistributions in binary form must reproduce the above copyright 
//_/_/    notice, this list of conditions and the following disclaimer in the 
//_/_/    documentation and/or other materials provided with the distribution.
//_/_/
//_/_/  - Neither the name of its copyright holders nor the names of its 
//_/_/    contributors may be used to endorse or promote products derived from 
//_/_/    this software without specific prior written permission.
//_/_/
//_/_/  - CADIA Clause: The license granted in and to the software under this 
//_/_/    agreement is a limited-use license. The software may not be used in 
//_/_/    furtherance of: (i) intentionally causing bodily injury or severe 
//_/_/    emotional harm to any person; (ii) invading the personal privacy or 
//_/_/    violating the human rights of any person; or (iii) committing or 
//_/_/    preparing for any act of war. 
//_/_/
//_/_/  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
//_/_/  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
//_/_/  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
//_/_/  PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
//_/_/  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
//_/_/  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
//_/_/  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
//_/_/  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
//_/_/  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
//_/_/  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
//_/_/  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//_/_/
//_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/
#if !defined(_MESSAGEPLAYER_H_)
#define _MESSAGEPLAYER_H_

#include "DataMessage.h"

namespace cmlabs {

struct ReplayIndexEntry {
	uint32 cid;
	uint64 serial;
	uint64 msgTime;
	uint32 size;
	char msgTriggerName[128];
	DataMessage* msg;
#if defined	ARCH_32
	DataMessage* dummy;
#endif
	bool reset() {
		memset(this, 0, sizeof(ReplayIndexEntry));
		cid = DATAMESSAGEINFOID;
		return true;
	}
	bool isValid() {
		return (cid == DATAMESSAGEINFOID);
	}
};


class MessagePlayer {
public:
	MessagePlayer();
	~MessagePlayer();

	bool initRead(const char* root, uint32 interval, bool rotate = false);
	bool initRead(const char* root, bool rotate = false);
	bool initWrite(const char* root, uint32 maxCount, uint64 maxSize);

	bool setSubTypeList(std::map<uint16, std::string>& subtypes);
	bool setSubContextList(std::map<uint16, std::string>& subcontexts);
	bool setComponentNameList(std::map<uint32, std::string>& names);
	bool writeMetadata();
	bool readMetadata();
	bool setSystemIDs(DataMessage* msg);

	bool exportToCSVFile(const char* filename, const char* separator = NULL);
	std::string exportToCSV(const char* separator = NULL);

	bool addMessage(const char* name, DataMessage* msg);
	DataMessage* waitForNextMessage(uint32 ms, const char* &triggerName);
	DataMessage* waitForNextMessage(uint32 ms, uint32 &msToNext);
	DataMessage* waitForNextMessage(uint32 ms, const char* &triggerName, uint32 &msToNext);
	DataMessage* waitForNextMessage(uint32 ms);
	std::string getCurrentTriggerName();

	bool flushToDisk();
	std::string printAllString(const char* format = NULL);

private:
	bool writeBuffer();
	bool readBuffer();

	uint32 interval;
	uint32 maxCount;
	uint64 maxSize;
	uint32 indexSize;
	bool rotate;
	std::string root;
	char* indexData;
	char currentTriggerName[MAXKEYNAMELEN + 1];
	std::string indexFile;
	uint32 indexFileSize;

	std::string metadataFile;
	std::string metadataTextFile;
	std::string csvFile;

	std::map<uint16, std::string> subtypeList;
	std::map<uint16, std::string> subcontextList;
	std::map<uint32, std::string> componentNameList;

	uint32 curStart;
	uint32 curWrite;
	uint32 curIndex;
	uint32 curCount;
	uint64 curSize;
	uint64 lastWriteTime;
	uint64 lastReceiveTime;
	uint64 lastReadTime;
	uint64 serial;
	uint64 nextMsgTime;
};

} // namespace cmlabs

#endif //_MESSAGEPLAYER_H_

