

//_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/
//_/_/
//_/_/  This file is part of CMSDK version 2.0.1.180405 (05 April 2018)
//_/_/  
//_/_/  developed by
//_/_/  
//_/_/  Communicative Machines Ltd. 
//_/_/    http://cmlabs.com/
//_/_/  Copyright (c) 2018
//_/_/  
//_/_/  in collaboration with 
//_/_/  
//_/_/  Icelandic Institute for Intelligent Machines ses
//_/_/    Menntavegur 1
//_/_/    101 Reykjavik
//_/_/    ICELAND
//_/_/    http://www.iiim.is/
//_/_/
//_/_/  and is released under the following
//_/_/  
//_/_/  EXTENDED BSD License
//_/_/
//_/_/  Redistribution and use in source and binary forms, with or without
//_/_/  modification, is permitted provided that the following conditions 
//_/_/  are met:
//_/_/
//_/_/  - Redistributions of source code must retain the above copyright 
//_/_/    and collaboration notice, this list of conditions and the following 
//_/_/    disclaimer.
//_/_/
//_/_/  - Redistributions in binary form must reproduce the above copyright 
//_/_/    notice, this list of conditions and the following disclaimer in the 
//_/_/    documentation and/or other materials provided with the distribution.
//_/_/
//_/_/  - Neither the name of its copyright holders nor the names of its 
//_/_/    contributors may be used to endorse or promote products derived from 
//_/_/    this software without specific prior written permission.
//_/_/
//_/_/  - CADIA Clause: The license granted in and to the software under this 
//_/_/    agreement is a limited-use license. The software may not be used in 
//_/_/    furtherance of: (i) intentionally causing bodily injury or severe 
//_/_/    emotional harm to any person; (ii) invading the personal privacy or 
//_/_/    violating the human rights of any person; or (iii) committing or 
//_/_/    preparing for any act of war. 
//_/_/
//_/_/  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
//_/_/  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
//_/_/  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
//_/_/  PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
//_/_/  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
//_/_/  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
//_/_/  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
//_/_/  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
//_/_/  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
//_/_/  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
//_/_/  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//_/_/
//_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/
#if !defined(_MOVINGAVERAGE_H_)
#define _MOVINGAVERAGE_H_

#include "Stats.h"
#include "PsyTime.h"

namespace cmlabs {

struct StatEntry {
	uint64 count;
	double value;
};

class MovingAverage {
public:
	static bool UnitTest();

	MovingAverage(uint32 binTimeMS = 100, uint16 binCount = 600);
	~MovingAverage();
	
	bool add(double val, uint32 count = 1, uint64 time = GetTimeNow());

	bool getThroughput(uint32 ms, double& valPerSec, double& countPerSec, uint64 now = GetTimeNow());
	bool getThroughputMulti(
		uint32 ms1, uint32 ms2, uint32 ms3,
		double& valPerSec1, double& countPerSec1,
		double& valPerSec2, double& countPerSec2,
		double& valPerSec3, double& countPerSec3, uint64 now = GetTimeNow() );

	bool getAverage(uint32 ms, double& val, uint64& count, uint64 now = GetTimeNow());
	bool getAverageMulti(
		uint32 ms1, uint32 ms2, uint32 ms3,
		double& avg1, uint64& count1,
		double& avg2, uint64& count2,
		double& avg3, uint64& count3, uint64 now = GetTimeNow());

	bool getSum(uint32 ms, double& val, uint64& count, uint64 now = GetTimeNow());
	bool getSumMulti(
		uint32 ms1, uint32 ms2, uint32 ms3,
		double& sum1, uint64& count1,
		double& sum2, uint64& count2,
		double& sum3, uint64& count3, uint64 now = GetTimeNow());
	
	bool getTotal(double& val, uint64& count);

protected:
	StatEntry* entries;
	utils::Mutex mutex;

	uint32 binTime;
	uint32 binCount;
	uint32 currentBin;
	uint64 currentBinStart;
	double totalValue;
	uint64 totalCount;
	// uint64 totalTime;

	bool shiftBins(uint64 now);
};


} // namespace cmlabs

#endif //_MOVINGAVERAGE_H_

