

//_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/
//_/_/
//_/_/  This file is part of CMSDK version 2.0.1.180405 (05 April 2018)
//_/_/  
//_/_/  developed by
//_/_/  
//_/_/  Communicative Machines Ltd. 
//_/_/    http://cmlabs.com/
//_/_/  Copyright (c) 2018
//_/_/  
//_/_/  in collaboration with 
//_/_/  
//_/_/  Icelandic Institute for Intelligent Machines ses
//_/_/    Menntavegur 1
//_/_/    101 Reykjavik
//_/_/    ICELAND
//_/_/    http://www.iiim.is/
//_/_/
//_/_/  and is released under the following
//_/_/  
//_/_/  EXTENDED BSD License
//_/_/
//_/_/  Redistribution and use in source and binary forms, with or without
//_/_/  modification, is permitted provided that the following conditions 
//_/_/  are met:
//_/_/
//_/_/  - Redistributions of source code must retain the above copyright 
//_/_/    and collaboration notice, this list of conditions and the following 
//_/_/    disclaimer.
//_/_/
//_/_/  - Redistributions in binary form must reproduce the above copyright 
//_/_/    notice, this list of conditions and the following disclaimer in the 
//_/_/    documentation and/or other materials provided with the distribution.
//_/_/
//_/_/  - Neither the name of its copyright holders nor the names of its 
//_/_/    contributors may be used to endorse or promote products derived from 
//_/_/    this software without specific prior written permission.
//_/_/
//_/_/  - CADIA Clause: The license granted in and to the software under this 
//_/_/    agreement is a limited-use license. The software may not be used in 
//_/_/    furtherance of: (i) intentionally causing bodily injury or severe 
//_/_/    emotional harm to any person; (ii) invading the personal privacy or 
//_/_/    violating the human rights of any person; or (iii) committing or 
//_/_/    preparing for any act of war. 
//_/_/
//_/_/  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
//_/_/  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
//_/_/  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
//_/_/  PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
//_/_/  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
//_/_/  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
//_/_/  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
//_/_/  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
//_/_/  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
//_/_/  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
//_/_/  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//_/_/
//_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/
#include "ControlMessage.h"

namespace cmlabs {

ControlMessage::ControlMessage(uint16 command, uint64 from, uint64 to, char* data, uint32 dataSize) {
	this->data = new char[sizeof(ControlMessageHeader) + dataSize];
	ControlMessageHeader* header = (ControlMessageHeader*) this->data;
	header->size = sizeof(ControlMessageHeader) + dataSize;
	header->cid = CONTROLMESSAGEID;
	header->command = command;
	header->time = GetTimeNow();
	header->from = from;
	header->to = to;
	if (dataSize > 0)
		memcpy(this->data+sizeof(ControlMessageHeader), data, dataSize);
}

ControlMessage::ControlMessage(char* data) {
	this->data = data;
}

ControlMessage::~ControlMessage() {
	data = NULL;
}

uint32 ControlMessage::getSize() {
	if (!VerObj(data)) return 0;
	return ((struct ControlMessageHeader*)data)->size;
}

uint16 ControlMessage::getCommand() {
	if (!VerObj(data)) return 0;
	return ((struct ControlMessageHeader*)data)->command;
}

uint64 ControlMessage::getTime() {
	if (!VerObj(data)) return 0;
	return ((struct ControlMessageHeader*)data)->time;
}

uint64 ControlMessage::getFrom() {
	if (!VerObj(data)) return 0;
	return ((struct ControlMessageHeader*)data)->from;
}

uint64 ControlMessage::getTo() {
	if (!VerObj(data)) return 0;
	return ((struct ControlMessageHeader*)data)->to;
}

uint16 ControlMessage::getStatus() {
	if (!VerObj(data)) return 0;
	return ((struct ControlMessageHeader*)data)->status;
}

bool ControlMessage::setStatus(uint16 status) {
	if (!VerObj(data)) return false;
	((ControlMessageHeader*)data)->status = status;
	return true;
}


} // namespace cmlabs
