

//_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/
//_/_/
//_/_/  This file is part of CMSDK version 2.0.1.180405 (05 April 2018)
//_/_/  
//_/_/  developed by
//_/_/  
//_/_/  Communicative Machines Ltd. 
//_/_/    http://cmlabs.com/
//_/_/  Copyright (c) 2018
//_/_/  
//_/_/  in collaboration with 
//_/_/  
//_/_/  Icelandic Institute for Intelligent Machines ses
//_/_/    Menntavegur 1
//_/_/    101 Reykjavik
//_/_/    ICELAND
//_/_/    http://www.iiim.is/
//_/_/
//_/_/  and is released under the following
//_/_/  
//_/_/  EXTENDED BSD License
//_/_/
//_/_/  Redistribution and use in source and binary forms, with or without
//_/_/  modification, is permitted provided that the following conditions 
//_/_/  are met:
//_/_/
//_/_/  - Redistributions of source code must retain the above copyright 
//_/_/    and collaboration notice, this list of conditions and the following 
//_/_/    disclaimer.
//_/_/
//_/_/  - Redistributions in binary form must reproduce the above copyright 
//_/_/    notice, this list of conditions and the following disclaimer in the 
//_/_/    documentation and/or other materials provided with the distribution.
//_/_/
//_/_/  - Neither the name of its copyright holders nor the names of its 
//_/_/    contributors may be used to endorse or promote products derived from 
//_/_/    this software without specific prior written permission.
//_/_/
//_/_/  - CADIA Clause: The license granted in and to the software under this 
//_/_/    agreement is a limited-use license. The software may not be used in 
//_/_/    furtherance of: (i) intentionally causing bodily injury or severe 
//_/_/    emotional harm to any person; (ii) invading the personal privacy or 
//_/_/    violating the human rights of any person; or (iii) committing or 
//_/_/    preparing for any act of war. 
//_/_/
//_/_/  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
//_/_/  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT 
//_/_/  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A 
//_/_/  PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER 
//_/_/  OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, 
//_/_/  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
//_/_/  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR 
//_/_/  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
//_/_/  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING 
//_/_/  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS 
//_/_/  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//_/_/
//_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/_/
#include "RequestClient.h"

namespace	cmlabs{

/////////////////////////////////////////////////////////////
// Request Reply
/////////////////////////////////////////////////////////////

bool RequestReply::isComplete() {
	switch (this->status) {
		case SUCCESS:
		case FAILED:
		case TIMEOUT:
		case LOCALERROR:
		case NETWORKERROR:
		case SERVERERROR:
			return true;
	}
	return false;
}

bool RequestReply::setStatus(RequestStatus status) {
	if (!mutex.enter(1000))
		return false;
	this->status = status;
	mutex.leave();
	return true;
}

RequestStatus RequestReply::getStatus() {
	RequestStatus res = LOCALERROR;
	if (!mutex.enter(1000))
		return res;
	res = this->status;
	mutex.leave();
	return res;
}


std::string RequestReply::getStatusText() {
	RequestStatus res = LOCALERROR;
	if (!mutex.enter(1000))
		return "Locking error";
	res = this->status;
	mutex.leave();
	switch (res) {
	case IDLE:
		return "Request idle";
	case QUEUED:
		return "Request queued";
	case SENT:
		return "Request sent";
	case SUCCESS:
		return "Request success";
	case FAILED:
		return "Request failed";
	case TIMEOUT:
		return "Request timed out";
	case LOCALERROR:
		return "Request local error";
	case NETWORKERROR:
		return "Request network error";
	case SERVERERROR:
		return "Request server error";
	default:
		return "Error getting status";
	}
}

RequestCallbackFunction RequestReply::getCallback() {
	RequestCallbackFunction res = NULL;
	if (!mutex.enter(1000))
		return res;
	res = this->callback;
	mutex.leave();
	return res;
}

bool RequestReply::setCallback(RequestCallbackFunction callback) {
	if (!mutex.enter(1000))
		return false;
	this->callback = callback;
	mutex.leave();
	return true;
}

bool RequestReply::giveRequestMessage(DataMessage* msg) {
	if (!mutex.enter(1000))
		return false;
	if (requestMsg)
		delete(requestMsg);
	requestMsg = msg;
	mutex.leave();
	return true;
}

bool RequestReply::setRequestMessageCopy(DataMessage* msg) {
	if (!mutex.enter(1000))
		return false;
	if (requestMsg)
		delete(requestMsg);
	if (msg)
		requestMsg = new DataMessage(*msg);
	else
		requestMsg = NULL;
	mutex.leave();
	return true;
}

DataMessage* RequestReply::getRequestMessageCopy() {
	DataMessage* msg = NULL;
	if (!mutex.enter(1000))
		return NULL;
	msg = new DataMessage(*requestMsg);
	mutex.leave();
	return msg;
}

DataMessage* RequestReply::peekRequestMessage() {
	return requestMsg;
}

bool RequestReply::giveReplyMessage(DataMessage* msg) {
	if (!mutex.enter(1000))
		return false;
	if (replyMsg)
		delete(replyMsg);
	replyMsg = msg;
	mutex.leave();
	return true;
}

bool RequestReply::setReplyMessageCopy(DataMessage* msg) {
	if (!mutex.enter(1000))
		return false;
	if (replyMsg)
		delete(replyMsg);
	if (msg)
		replyMsg = new DataMessage(*msg);
	else
		replyMsg = NULL;
	mutex.leave();
	return true;
}

DataMessage* RequestReply::getReplyMessageCopy() {
	DataMessage* msg = NULL;
	if (!mutex.enter(1000))
		return NULL;
	msg = new DataMessage(*replyMsg);
	mutex.leave();
	return msg;
}

DataMessage* RequestReply::peekReplyMessage() {
	return replyMsg;
}

bool RequestReply::replyToRequest(DataMessage* msg, RequestStatus status) {
	if (!msg) return false;
	if (!mutex.enter(1000))
		return false;
	if (replyMsg)
		delete(replyMsg);
	replyMsg = msg;
	this->status = status;
	finishTime = GetTimeNow();
	semaphore.signal();

	//if (customRef == 2) {
	//	printf("--- [%u] --- Long request %.3fms \n", msg->getFrom(), ((double)finishTime-startTime)/1000.0);
	//	fflush(stdout);
	//}
	//else if (customRef == 1) {
	//	printf("--- [%u] --- Short request %.3fms \n", msg->getFrom(), ((double)finishTime-startTime)/1000.0);
	//	fflush(stdout);
	//}

	mutex.leave();
	return true;
}

uint64 RequestReply::getRequestDuration() {
	if (finishTime)
		return finishTime-startTime;
	else
		return 0;
}

uint32 RequestReply::getRequestDurationMS() {
	if (finishTime)
		return (uint32)((finishTime-startTime)/1000);
	else
		return 0;
}

RequestStatus RequestReply::waitForResult(uint32 timeoutMS) {
	RequestStatus result;
	if (!mutex.enter(200, __FUNCTION__))
		return LOCALERROR;

	//if ( (status == SUCCESS) || (status == FAILED) || (status == TIMEOUT) || (status == LOCALERROR) || (status == NETWORKERROR) || (status == SERVERERROR) ) {
	if (isComplete()) {
		result = status;
		mutex.leave();
		return result;
	}

	mutex.leave();
	if (semaphore.wait(timeoutMS)) {
		if (!mutex.enter(200, __FUNCTION__))
			return LOCALERROR;
		//if ( (status == SUCCESS) || (status == FAILED) || (status == TIMEOUT) || (status == LOCALERROR) || (status == NETWORKERROR) || (status == SERVERERROR) ) {
		if (isComplete()) {
			result = status;
			mutex.leave();
			return result;
		}
		mutex.leave();
	}
	return TIMEOUT;
}

DataMessage* RequestReply::waitForMessage(uint32 timeoutMS, bool takeMessage) {
	DataMessage* msg;
	if (!mutex.enter(200, __FUNCTION__))
		return NULL;

	if (replyMsg) {
		msg = replyMsg;
		if (takeMessage)
			replyMsg = NULL;
		mutex.leave();
		return msg;
	}

	mutex.leave();
	if (semaphore.wait(timeoutMS)) {
		if (!mutex.enter(200, __FUNCTION__))
			return NULL;
		if (replyMsg) {
			msg = replyMsg;
			if (takeMessage)
				replyMsg = NULL;
			mutex.leave();
			return msg;
		}
		mutex.leave();
	}
	return NULL;
}


/////////////////////////////////////////////////////////////
// Request Client
/////////////////////////////////////////////////////////////

RequestClient::RequestClient() {
	lastRefID = 0;
	sentCount = 0;
	receivedCount = 0;
	shouldContinue = true;
	isRunning = false;
	channel = NULL;
	manager = new NetworkManager();

	if (!ThreadManager::CreateThread(RequestClientRun, this, threadID, 0)) {
		shouldContinue = false;
	}
}

RequestClient::~RequestClient() {
	conMutex.enter(3000, __FUNCTION__);
	stop();

	std::map<uint64, RequestReply*>::iterator i, e;
	mutex.enter(200, __FUNCTION__);
	for (i = requestMap.begin(), e = requestMap.end(); i != e; i++)
		delete(i->second);
	requestMap.clear();
	mutex.leave();
	conMutex.leave();

	delete(manager);
}

bool RequestClient::addGateway(uint32 id, std::string addr, uint16 port, uint8 encryption) {
	RequestGatewayConnection con;
	con.clear();
	con.id = id;
	con.addr = addr;
	con.port = port;
	con.encryption = encryption;
	connections.push_back(con);
	return true;
}

RequestReply* RequestClient::postRequest(DataMessage *msg) {
	if (!msg)
		return NULL;
	if (!mutex.enter(200, __FUNCTION__))
		return NULL;
	msg->setReference(++lastRefID);
	
	RequestReply* reply = new RequestReply();
	reply->setStatus(QUEUED);
	reply->giveRequestMessage(msg);
	reply->clientRef = lastRefID;
	requestMap[lastRefID] = reply;
	outQ.add(reply);
	mutex.leave();
	return reply;
}

bool RequestClient::postRequest(DataMessage *msg, RequestCallbackFunction callback, uint32 timeoutMS) {
	// ##############
	return false;
}

bool RequestClient::finishRequest(RequestReply* reply) {
	if (!reply || !reply->clientRef) return false;
	if (!mutex.enter(200, __FUNCTION__))
		return false;

	std::map<uint64, RequestReply*>::iterator i = requestMap.find(reply->clientRef);
	if (i == requestMap.end())
		return false;
	if (i->second) {
		delete i->second;
		requestMap.erase(i);
		//if (i->second->isComplete()) {
		//	delete i->second;
		//	requestMap.erase(i);
		//}
		//else
		//	i->second->isInUse = false;
	}
	mutex.leave();
	return true;
}


bool RequestClient::receiveNetworkEvent(NetworkEvent* evt, NetworkChannel* channel, uint64 conid) {
	uint64 now = GetTimeNow();
	std::list<RequestGatewayConnection>::iterator i, e = connections.end();
	if (!mutex.enter(200, __FUNCTION__))
		return false;
	i = connections.begin();
	while (i != e) {
		if ((*i).conID == conid) {
			switch (evt->type) {
			case NETWORKEVENT_DISCONNECT:
				(*i).lastFailTime = now;
				(*i).conID = 0;
				break;
			case NETWORKEVENT_RECONNECT:
				(*i).lastFailTime = 0;
				break;
			default:
				break;
			}
		}
		i++;
	}
	mutex.leave();
	return true;
}

bool RequestClient::receiveMessage(DataMessage* msg, NetworkChannel* channel, uint64 conid) {
	if (!mutex.enter(200, __FUNCTION__))
		return false;
	uint64 ref = msg->getReference();
	std::map<uint64,RequestReply*>::iterator i = requestMap.find(ref);

	if (i == requestMap.end()) {
		// #############
		delete(msg);
		return true;
	}
	RequestReply* reply = i->second;
	reply->replyToRequest(msg, SUCCESS);
	if (!reply->isInUse) {
		requestMap.erase(i);
		delete reply;
	}
	receivedCount++;
	mutex.leave();
	return true;
}

bool RequestClient::receiveHTTPReply(HTTPReply* reply, HTTPRequest* req, NetworkChannel* channel, uint64 conid) {
	return false;
}


bool RequestClient::sendMessageToGateway(DataMessage* msg, RequestGatewayConnection& con) {

	if (!con.conID || !channel) {
		if (!channel)
			channel = manager->createTCPConnection(con.addr.c_str(), con.port, con.encryption, PROTOCOL_MESSAGE, true, true, con.id, this, con.conID, con.location);
		else
			con.conID = channel->createTCPConnection(con.addr.c_str(), con.port, con.encryption, PROTOCOL_MESSAGE, true, true, con.location);
		if (!channel || !con.conID) {
			con.lastFailTime = GetTimeNow();
			return false;
		}
		DataMessage* msg = new DataMessage();
		msg->setString("URI", "ClientConnect");
		if (!channel->sendMessage(msg, con.conID)) {
			// clearly not working, disconnect and try again later
			channel->endConnection(con.conID);
			con.conID = 0;
			con.lastFailTime = GetTimeNow();
			return false;
		}
		con.lastConTime = GetTimeNow();
	}
	if (!channel->sendMessage(msg, con.conID)) {
		// clearly not working, disconnect and try again later
		channel->endConnection(con.conID);
		con.conID = 0;
		con.lastFailTime = GetTimeNow();
		return false;
	}
	con.lastFailTime = 0;
	return true;
}


bool RequestClient::run() {
	RequestReply* reply;
	isRunning = true;

	// should regularly test for old replies to be deleted...
	// ###############

	while (shouldContinue) {
		//if (!mutex.enter(200, __FUNCTION__))
		//	continue;
		if (!channel && connections.size()) {
			conMutex.enter(3000, __FUNCTION__);
			DataMessage* msg = new DataMessage();
			msg->setString("URI", "ClientStatus");
			std::list<RequestGatewayConnection>::iterator iCon, eCon = connections.end();
			iCon = connections.begin();
			while (iCon != eCon) {
				if (sendMessageToGateway(msg, (*iCon)))
					break;
				iCon++;
			}
			delete msg;
			conMutex.leave();
		}

		while (reply = outQ.waitForAndTakeFirst(50))
			sendRequest(reply);

		//mutex.leave();
	}
	isRunning = false;
	return true;
}

bool RequestClient::isConnected() {
	if (!conMutex.enter(3000, __FUNCTION__))
		return false;
	std::list<RequestGatewayConnection>::iterator iCon, eCon = connections.end();
	iCon = connections.begin();
	while (iCon != eCon) {
		//printf("RUN0\n");fflush(stdout);
		if ((*iCon).conID && channel->isConnected((*iCon).conID)) {
			conMutex.leave();
			return true;
		}
		iCon++;
	}
	bool res = false;
	// None were connected, try to connect
	DataMessage* msg = new DataMessage();
	msg->setString("URI", "ClientStatus");
	iCon = connections.begin();
	while (iCon != eCon) {
		if (sendMessageToGateway(msg, (*iCon)))
			break;
		iCon++;
	}
	conMutex.leave();
	return res;
}

bool RequestClient::sendRequest(RequestReply* reply) {
	if (!reply)
		return false;

	uint64 now = GetTimeNow();
	conMutex.enter(3000, __FUNCTION__);
	std::list<RequestGatewayConnection>::iterator iCon, eCon = connections.end();
	iCon = connections.begin();
	while (iCon != eCon) {
		if (sendMessageToGateway(reply->peekRequestMessage(), (*iCon))) {
			reply->setStatus(SENT);
			sentCount++;
			conMutex.leave();
			return true;
		}
		iCon++;
	}
	conMutex.leave();
	return false;
}

THREAD_RET THREAD_FUNCTION_CALL RequestClientRun(THREAD_ARG arg) {
	if (arg == NULL) thread_ret_val(1);
	thread_ret_val((int)(((RequestClient*)arg)->run() ? 0 : 1));
}











TestRequestClient::TestRequestClient(bool isLong) : RequestClient() {
	uint32 myThreadID;
	ThreadManager::CreateThread(TestRequestClientRun, this, myThreadID, 0);
}


bool TestRequestClient::testRun() {
	//DataMessage* reply, *msg;
	while (shouldContinue) {
		utils::Sleep(20);
	}
	return true;
}

THREAD_RET THREAD_FUNCTION_CALL TestRequestClientRun(THREAD_ARG arg) {
	if (arg == NULL) thread_ret_val(1);
	thread_ret_val((int)(((TestRequestClient*)arg)->testRun() ? 0 : 1));
}



std::string RequestClient::GetJSONReplyParameter(const char* json, const char* name) {

	jsmn_parser parser;
	jsmntok_t* tokens = new jsmntok_t[1024];
	int tokenCount = 0;
	const char* val;
	int valSize;
	jsmn_init(&parser);
	if ((tokenCount = jsmn_parse(&parser, json, strlen(json), tokens, 1024)) <= 0) {
		// No JSON found
		delete[] tokens;
		return "";
	}
	if (!(val = GetJSONValue(tokens, tokenCount, json, name, JSMN_UNDEFINED, valSize))) {
		// JSON key not found
		delete[] tokens;
		return "";
	}
	delete[] tokens;
	std::string value = std::string(val, valSize);
	return value.c_str();
}

DataMessage* RequestClient::SendRequestAndWaitForReply(RequestClient* client, DataMessage* msg, int timeoutMS) {
	if (!client->isConnected())
		return NULL;

	DataMessage *replyMsg;
	RequestReply* reply;
	if (reply = client->postRequest(msg)) {
		if (replyMsg = reply->waitForMessage(timeoutMS, true)) {
			client->finishRequest(reply);
			//delete(reply);
			return replyMsg;
		}
		client->finishRequest(reply);
		//delete(reply);
	}
	return NULL;
}


std::string RequestClient::SendRequestAndWaitForJSON(RequestClient* client, DataMessage* msg, int timeoutMS) {
	if (!client->isConnected())
		return "";

	DataMessage *replyMsg = SendRequestAndWaitForReply(client, msg, timeoutMS);
	if (!replyMsg)
		return "";
	const char* json = replyMsg->getString("JSON");
	if (!json) {
		delete(replyMsg);
		return "";
	}

	std::string jsonReply = json;
	delete(replyMsg);
	return jsonReply;
}

DataMessage* RequestClient::CreateRequestMessage(uint8 operation, const char* req,
	const char* key1, const char* val1,
	const char* key2, const char* val2,
	const char* key3, const char* val3,
	const char* key4, const char* val4,
	const char* key5, const char* val5,
	const char* key6, const char* val6,
	const char* key7, const char* val7,
	const char* key8, const char* val8,
	const char* key9, const char* val9,
	const char* key10, const char* val10,
	const char* key11, const char* val11,
	const char* key12, const char* val12,
	const char* key13, const char* val13,
	const char* key14, const char* val14,
	const char* key15, const char* val15,
	const char* key16, const char* val16,
	const char* key17, const char* val17,
	const char* key18, const char* val18,
	const char* key19, const char* val19,
	const char* key20, const char* val20
	) {
	DataMessage* msg = new DataMessage();
	msg->setInt("HTTP_OPERATION", operation);
	msg->setString("URI", req);
	if (key1) msg->setString(key1, val1);
	if (key2) msg->setString(key2, val2);
	if (key3) msg->setString(key3, val3);
	if (key4) msg->setString(key4, val4);
	if (key5) msg->setString(key5, val5);
	if (key6) msg->setString(key6, val6);
	if (key7) msg->setString(key7, val7);
	if (key8) msg->setString(key8, val8);
	if (key9) msg->setString(key9, val9);
	if (key10) msg->setString(key10, val10);
	if (key11) msg->setString(key11, val11);
	if (key12) msg->setString(key12, val12);
	if (key13) msg->setString(key13, val13);
	if (key14) msg->setString(key14, val14);
	if (key15) msg->setString(key15, val15);
	if (key16) msg->setString(key16, val16);
	if (key17) msg->setString(key17, val17);
	if (key18) msg->setString(key18, val18);
	if (key19) msg->setString(key19, val19);
	if (key20) msg->setString(key20, val20);
	return msg;
}





bool RequestClient::TestServerLogin(const char* address, uint16 port, const char* username, const char* password, const char* reqAfterLogin) {

	RequestClient* reqClient = new RequestClient();
	//	api->logPrint(1, "Connecting to Face Server on %s:%u...", addr.c_str(), port, username.c_str());
	if (!reqClient->addGateway(1, address, port, NOENC)) {
		LogPrint(0, 0, 1, "Could not connect to Face Server on %s:%u...", address, port);
		delete(reqClient);
		return false;
	}

	std::string json;
	if (!(json = RequestClient::SendRequestAndWaitForJSON(
		reqClient,
		RequestClient::CreateRequestMessage(HTTP_POST, "authenticate", "email", username, "password", password, "clientid", "1", "locationid", "1"),
		5000)).length()) {
		LogPrint(0, 0, 1, "Couldn't login to server account...");
		return false;
	}
	std::string token = RequestClient::GetJSONReplyParameter(json.c_str(), "token");
	if (!token.length()) {
		LogPrint(0, 0, 1, "Client login failed: %s...", RequestClient::GetJSONReplyParameter(json.c_str(), "text").c_str());
		return false;
	}

	LogPrint(0, 0, 1, "Successfully logged in with token: %s...", token.c_str());

	if (!reqClient->isConnected()) {
		LogPrint(0, 0, 1, "Couldn't connect to server...");
		return false;
	}

	return true;
}


}
